from datetime import date, datetime, time
from unittest.mock import Mock

from octue import exceptions
from octue.mixins.filterable import Filterable
from octue.resources.label import LabelSet
from octue.resources.tag import TagDict
from tests.base import BaseTestCase


class FilterableSubclass(Filterable):
    def __init__(self, **kwargs):
        for key, value in kwargs.items():
            setattr(self, key, value)


class TestFilterable(BaseTestCase):
    def test_error_raised_when_invalid_filter_name_received(self):
        """ Ensure an error is raised when an invalid filter name is provided. """
        with self.assertRaises(exceptions.InvalidInputException):
            FilterableSubclass().satisfies(invalid_filter_name=None)

    def test_error_raised_when_non_existent_attribute_name_received(self):
        """ Ensure an error is raised when a non-existent attribute name is used in the filter name. """
        with self.assertRaises(AttributeError):
            FilterableSubclass().satisfies(boogaloo__is_a_dance=True)

    def test_error_raised_when_valid_but_non_existent_filter_name_received(self):
        """ Ensure an error is raised when a valid but non-existent filter name is received. """
        with self.assertRaises(exceptions.InvalidInputException):
            FilterableSubclass(age=23).satisfies(age__is_secret=True)

    def test_error_raised_when_attribute_type_has_no_filters_defined(self):
        """Ensure an error is raised when a filter for an attribute whose type doesn't have any filters defined is
        received.
        """
        with self.assertRaises(exceptions.InvalidInputException):
            FilterableSubclass(age=lambda: None).satisfies(age__equals=True)

    def test_error_raised_if_more_than_one_filter_is_provided(self):
        """Test that an error is raised if more than one filter is provided to the satisfies method."""
        with self.assertRaises(ValueError):
            FilterableSubclass(age=23).satisfies(age__equals=23, age__equals__gt=20)

    def test_equals_filter_shortcut(self):
        """Test that the shortcut for simple equals filters works (e.g. `a=7` instead of `a__equals=7`)."""
        self.assertTrue(FilterableSubclass(age=23).satisfies(age=23))
        self.assertFalse(FilterableSubclass(age=23).satisfies(age=32))

    def test_equals_filter_shortcut_with_nested_attributes(self):
        """Test that the shortcut for simple equals filters works for nested attributes e.g. `a__b=7` instead of
        `a__b__equals=7`.
        """
        inner_mock = FilterableSubclass(b=3)
        outer_mock = FilterableSubclass(a=inner_mock)
        filterable_thing = FilterableSubclass(name=outer_mock)
        self.assertTrue(filterable_thing.satisfies(name__a__b=3))

    def test_bool_filters(self):
        """ Test that the boolean filters work as expected. """
        filterable_thing = FilterableSubclass(is_alive=True)
        self.assertTrue(filterable_thing.satisfies(is_alive__is=True))
        self.assertFalse(filterable_thing.satisfies(is_alive__is=False))
        self.assertTrue(filterable_thing.satisfies(is_alive__is_not=False))
        self.assertFalse(filterable_thing.satisfies(is_alive__is_not=True))
        self.assertTrue(filterable_thing.satisfies(is_alive__equals=True))
        self.assertFalse(filterable_thing.satisfies(is_alive__equals=False))
        self.assertTrue(filterable_thing.satisfies(is_alive__not_equals=False))
        self.assertFalse(filterable_thing.satisfies(is_alive__not_equals=True))

    def test_str_filters(self):
        """ Test that the string filters work as expected. """
        filterable_thing = FilterableSubclass(name="Michael")
        self.assertTrue(filterable_thing.satisfies(name__icontains="m"))
        self.assertFalse(filterable_thing.satisfies(name__icontains="d"))
        self.assertTrue(filterable_thing.satisfies(name__not_icontains="d"))
        self.assertFalse(filterable_thing.satisfies(name__not_icontains="m"))
        self.assertTrue(filterable_thing.satisfies(name__contains="M"))
        self.assertFalse(filterable_thing.satisfies(name__contains="d"))
        self.assertTrue(filterable_thing.satisfies(name__ends_with="l"))
        self.assertFalse(filterable_thing.satisfies(name__ends_with="M"))
        self.assertTrue(filterable_thing.satisfies(name__not_ends_with="M"))
        self.assertFalse(filterable_thing.satisfies(name__not_ends_with="l"))
        self.assertTrue(filterable_thing.satisfies(name__starts_with="M"))
        self.assertFalse(filterable_thing.satisfies(name__starts_with="l"))
        self.assertTrue(filterable_thing.satisfies(name__not_starts_with="l"))
        self.assertFalse(filterable_thing.satisfies(name__not_starts_with="M"))
        self.assertTrue(filterable_thing.satisfies(name__equals="Michael"))
        self.assertFalse(filterable_thing.satisfies(name__equals="Clive"))
        self.assertTrue(filterable_thing.satisfies(name__not_equals="Clive"))
        self.assertFalse(filterable_thing.satisfies(name__not_equals="Michael"))
        self.assertTrue(filterable_thing.satisfies(name__iequals="michael"))
        self.assertFalse(filterable_thing.satisfies(name__iequals="James"))
        self.assertTrue(filterable_thing.satisfies(name__not_iequals="James"))
        self.assertFalse(filterable_thing.satisfies(name__not_iequals="michael"))
        self.assertTrue(filterable_thing.satisfies(name__is="Michael"))
        self.assertFalse(filterable_thing.satisfies(name__is="Clive"))
        self.assertTrue(filterable_thing.satisfies(name__is_not="Clive"))
        self.assertFalse(filterable_thing.satisfies(name__is_not="Michael"))
        self.assertTrue(filterable_thing.satisfies(name__lt="Noel"))
        self.assertFalse(filterable_thing.satisfies(name__lt="Harry"))
        self.assertTrue(filterable_thing.satisfies(name__lte="Michael"))
        self.assertFalse(filterable_thing.satisfies(name__lte="Harry"))
        self.assertTrue(filterable_thing.satisfies(name__gt="Clive"))
        self.assertFalse(filterable_thing.satisfies(name__gt="Noel"))
        self.assertTrue(filterable_thing.satisfies(name__gte="Michael"))
        self.assertFalse(filterable_thing.satisfies(name__gte="Noel"))
        self.assertTrue(filterable_thing.satisfies(name__in_range=("Amy", "Zoe")))
        self.assertFalse(filterable_thing.satisfies(name__in_range=("Noel", "Peter")))
        self.assertTrue(filterable_thing.satisfies(name__not_in_range=("Noel", "Peter")))
        self.assertFalse(filterable_thing.satisfies(name__not_in_range=("Amy", "Zoe")))

    def test_none_filters(self):
        """ Test that the None filters work as expected. """
        filterable_thing = FilterableSubclass(owner=None)
        self.assertTrue(filterable_thing.satisfies(owner__is=None))
        self.assertFalse(filterable_thing.satisfies(owner__is=True))
        self.assertTrue(filterable_thing.satisfies(owner__is_not=True))
        self.assertFalse(filterable_thing.satisfies(owner__is_not=None))
        self.assertTrue(filterable_thing.satisfies(owner__equals=None))
        self.assertFalse(filterable_thing.satisfies(owner__equals=True))
        self.assertTrue(filterable_thing.satisfies(owner__not_equals=True))
        self.assertFalse(filterable_thing.satisfies(owner__not_equals=None))

    def test_number_filters_with_integers_and_floats(self):
        """ Test that the number filters work as expected for integers and floats. """
        for age in (5, 5.2):
            filterable_thing = FilterableSubclass(age=age)
            self.assertTrue(filterable_thing.satisfies(age__equals=age))
            self.assertFalse(filterable_thing.satisfies(age__equals=63))
            self.assertTrue(filterable_thing.satisfies(age__not_equals=63))
            self.assertFalse(filterable_thing.satisfies(age__not_equals=age))
            self.assertTrue(filterable_thing.satisfies(age__lt=6))
            self.assertFalse(filterable_thing.satisfies(age__lt=0))
            self.assertTrue(filterable_thing.satisfies(age__lte=age))
            self.assertFalse(filterable_thing.satisfies(age__lte=0))
            self.assertTrue(filterable_thing.satisfies(age__gt=4))
            self.assertFalse(filterable_thing.satisfies(age__gt=63))
            self.assertTrue(filterable_thing.satisfies(age__gte=age))
            self.assertFalse(filterable_thing.satisfies(age__gte=63))
            self.assertTrue(filterable_thing.satisfies(age__is=age))
            self.assertFalse(filterable_thing.satisfies(age__is=63))
            self.assertTrue(filterable_thing.satisfies(age__is_not=63))
            self.assertFalse(filterable_thing.satisfies(age__is_not=age))
            self.assertTrue(filterable_thing.satisfies(age__in_range=(0, 10)))
            self.assertFalse(filterable_thing.satisfies(age__in_range=(0, 3)))
            self.assertTrue(filterable_thing.satisfies(age__not_in_range=(0, 3)))
            self.assertFalse(filterable_thing.satisfies(age__not_in_range=(0, 10)))

    def test_iterable_filters(self):
        """ Test that the iterable filters work as expected with lists, sets, and tuples. """
        for iterable in ([1, 2, 3], {1, 2, 3}, (1, 2, 3)):
            filterable_thing = FilterableSubclass(iterable=iterable)
            self.assertTrue(filterable_thing.satisfies(iterable__contains=1))
            self.assertFalse(filterable_thing.satisfies(iterable__contains=5))
            self.assertTrue(filterable_thing.satisfies(iterable__not_contains=5))
            self.assertFalse(filterable_thing.satisfies(iterable__not_contains=1))
            self.assertTrue(filterable_thing.satisfies(iterable__is=iterable))
            self.assertFalse(filterable_thing.satisfies(iterable__is=None))
            self.assertTrue(filterable_thing.satisfies(iterable__is_not=None))
            self.assertFalse(filterable_thing.satisfies(iterable__is_not=iterable))

    def test_label_set_filters(self):
        """ Test the filters for Labelset. """
        filterable_thing = FilterableSubclass(iterable=LabelSet({"fred", "charlie"}))
        self.assertTrue(filterable_thing.satisfies(iterable__any_label_contains="a"))
        self.assertFalse(filterable_thing.satisfies(iterable__any_label_contains="z"))
        self.assertTrue(filterable_thing.satisfies(iterable__not_any_label_contains="z"))
        self.assertFalse(filterable_thing.satisfies(iterable__not_any_label_contains="a"))
        self.assertTrue(filterable_thing.satisfies(iterable__any_label_starts_with="f"))
        self.assertFalse(filterable_thing.satisfies(iterable__any_label_starts_with="e"))
        self.assertTrue(filterable_thing.satisfies(iterable__any_label_ends_with="e"))
        self.assertFalse(filterable_thing.satisfies(iterable__any_label_ends_with="i"))
        self.assertTrue(filterable_thing.satisfies(iterable__not_any_label_starts_with="e"))
        self.assertFalse(filterable_thing.satisfies(iterable__not_any_label_starts_with="f"))
        self.assertTrue(filterable_thing.satisfies(iterable__not_any_label_ends_with="i"))
        self.assertFalse(filterable_thing.satisfies(iterable__not_any_label_ends_with="e"))

    def test_datetime_filters(self):
        my_datetime = datetime(2000, 1, 1)
        filterable_thing = FilterableSubclass(timestamp=my_datetime)
        self.assertTrue(filterable_thing.satisfies(timestamp__equals=my_datetime))
        self.assertFalse(filterable_thing.satisfies(timestamp__equals=datetime(2, 2, 2)))
        self.assertTrue(filterable_thing.satisfies(timestamp__not_equals=datetime(2, 2, 2)))
        self.assertFalse(filterable_thing.satisfies(timestamp__not_equals=my_datetime))
        self.assertTrue(filterable_thing.satisfies(timestamp__is=my_datetime))
        self.assertFalse(filterable_thing.satisfies(timestamp__is=datetime(2, 2, 2)))
        self.assertTrue(filterable_thing.satisfies(timestamp__is_not=datetime(2, 2, 2)))
        self.assertFalse(filterable_thing.satisfies(timestamp__is_not=my_datetime))
        self.assertTrue(filterable_thing.satisfies(timestamp__gt=datetime(1900, 1, 2)))
        self.assertFalse(filterable_thing.satisfies(timestamp__gt=datetime(3000, 1, 2)))
        self.assertTrue(filterable_thing.satisfies(timestamp__gte=my_datetime))
        self.assertFalse(filterable_thing.satisfies(timestamp__gte=datetime(3000, 1, 2)))
        self.assertTrue(filterable_thing.satisfies(timestamp__lt=datetime(3000, 1, 2)))
        self.assertFalse(filterable_thing.satisfies(timestamp__lt=datetime(1990, 1, 2)))
        self.assertTrue(filterable_thing.satisfies(timestamp__lte=my_datetime))
        self.assertFalse(filterable_thing.satisfies(timestamp__lte=datetime(1900, 1, 2)))
        self.assertTrue(filterable_thing.satisfies(timestamp__in_range=(datetime(1900, 1, 2), datetime(3000, 1, 2))))
        self.assertFalse(filterable_thing.satisfies(timestamp__in_range=(datetime(2100, 1, 2), datetime(3000, 1, 2))))
        self.assertTrue(
            filterable_thing.satisfies(timestamp__not_in_range=(datetime(2100, 1, 2), datetime(3000, 1, 2)))
        )
        self.assertFalse(
            filterable_thing.satisfies(timestamp__not_in_range=(datetime(1900, 1, 2), datetime(3000, 1, 2)))
        )
        self.assertTrue(filterable_thing.satisfies(timestamp__year_equals=2000))
        self.assertFalse(filterable_thing.satisfies(timestamp__year_equals=3000))
        self.assertTrue(filterable_thing.satisfies(timestamp__year_in={2000, 3000, 4000}))
        self.assertFalse(filterable_thing.satisfies(timestamp__year_in={3000, 4000}))
        self.assertTrue(filterable_thing.satisfies(timestamp__month_equals=1))
        self.assertFalse(filterable_thing.satisfies(timestamp__month_equals=9))
        self.assertTrue(filterable_thing.satisfies(timestamp__month_in={1, 2, 3}))
        self.assertFalse(filterable_thing.satisfies(timestamp__month_in={2, 3}))
        self.assertTrue(filterable_thing.satisfies(timestamp__day_equals=1))
        self.assertFalse(filterable_thing.satisfies(timestamp__day_equals=2))
        self.assertTrue(filterable_thing.satisfies(timestamp__day_in={1, 2, 3}))
        self.assertFalse(filterable_thing.satisfies(timestamp__day_in={2, 3}))
        self.assertTrue(filterable_thing.satisfies(timestamp__weekday_equals=5))
        self.assertFalse(filterable_thing.satisfies(timestamp__weekday_equals=3))
        self.assertTrue(filterable_thing.satisfies(timestamp__weekday_in={5, 6, 7}))
        self.assertFalse(filterable_thing.satisfies(timestamp__weekday_in={6, 7}))
        self.assertTrue(filterable_thing.satisfies(timestamp__iso_weekday_equals=6))
        self.assertFalse(filterable_thing.satisfies(timestamp__iso_weekday_equals=4))
        self.assertTrue(filterable_thing.satisfies(timestamp__iso_weekday_in={5, 6, 7}))
        self.assertFalse(filterable_thing.satisfies(timestamp__iso_weekday_in={7, 8}))
        self.assertTrue(filterable_thing.satisfies(timestamp__time_equals=time(0, 0, 0)))
        self.assertFalse(filterable_thing.satisfies(timestamp__time_equals=time(1, 2, 3)))
        self.assertTrue(filterable_thing.satisfies(timestamp__hour_equals=0))
        self.assertFalse(filterable_thing.satisfies(timestamp__hour_equals=1))
        self.assertTrue(filterable_thing.satisfies(timestamp__hour_in={0, 1, 2}))
        self.assertFalse(filterable_thing.satisfies(timestamp__hour_in={1, 2}))
        self.assertTrue(filterable_thing.satisfies(timestamp__minute_equals=0))
        self.assertFalse(filterable_thing.satisfies(timestamp__minute_equals=1))
        self.assertTrue(filterable_thing.satisfies(timestamp__minute_in={0, 1, 2}))
        self.assertFalse(filterable_thing.satisfies(timestamp__minute_in={1, 2}))
        self.assertTrue(filterable_thing.satisfies(timestamp__second_equals=0))
        self.assertFalse(filterable_thing.satisfies(timestamp__second_equals=1))
        self.assertTrue(filterable_thing.satisfies(timestamp__second_in={0, 1, 2}))
        self.assertFalse(filterable_thing.satisfies(timestamp__second_in={1, 2}))
        self.assertTrue(filterable_thing.satisfies(timestamp__in_date_range=(date(1000, 1, 4), date(3000, 7, 10))))
        self.assertFalse(filterable_thing.satisfies(timestamp__in_date_range=(date(2000, 1, 4), date(3000, 7, 10))))
        self.assertTrue(filterable_thing.satisfies(timestamp__in_time_range=(time(0, 0, 0), time(13, 2, 22))))
        self.assertFalse(filterable_thing.satisfies(timestamp__in_time_range=(time(0, 0, 1), time(13, 2, 22))))

    def test_filtering_different_attributes_on_same_instance(self):
        """ Ensure all filterable attributes on an instance can be checked for filter satisfaction. """
        filterable_thing = FilterableSubclass(name="Fred", is_alive=True, iterable={1, 2, 3}, age=5.2, owner=None)
        self.assertTrue(filterable_thing.satisfies(name__icontains="f"))
        self.assertTrue(filterable_thing.satisfies(name__not_icontains="j"))
        self.assertFalse(filterable_thing.satisfies(is_alive__is=False))
        self.assertTrue(filterable_thing.satisfies(iterable__contains=3))
        self.assertTrue(filterable_thing.satisfies(age__equals=5.2))
        self.assertTrue(filterable_thing.satisfies(age__not_equals=5))
        self.assertTrue(filterable_thing.satisfies(owner__is=None))

    def test_filtering_with_nested_attributes(self):
        """Test that Filterable subclasses can be checked for satisfaction of a filter of nested attributes."""
        inner_mock = Mock(b=3)
        outer_mock = Mock(a=inner_mock)
        filterable_thing = FilterableSubclass(name=outer_mock)
        self.assertTrue(filterable_thing.satisfies(name__a__b__equals=3))

    def test_filtering_with_nested_attributes_ending_in_dictionary_key(self):
        """Test that Filterable subclasses can be checked for satisfaction of a filter of nested attributes that ends
        with a dictionary key.
        """
        filterable_thing = FilterableSubclass(name={"first": "Joe", "last": "Bloggs"})
        self.assertTrue(filterable_thing.satisfies(name__first__equals="Joe"))
        self.assertTrue(filterable_thing.satisfies(name__last__equals="Bloggs"))

    def test_tag_dict_filters(self):
        """Test some filters that apply to a TagDict. These should behave just the same as for a dictionary."""
        filterable_thing = FilterableSubclass(tags=TagDict({"first": "Joe", "middle": "Horatio", "last": "Bloggs"}))
        self.assertTrue(filterable_thing.satisfies(tags__last__lt="Kevin"))
        self.assertFalse(filterable_thing.satisfies(tags__middle__is="Boratio"))



import logging
import logging.handlers
from urllib.parse import urlparse


def create_formatter(logging_metadata):
    """Create a log formatter from the given logging metadata that delimits the context fields with space-padded
    pipes and encapsulates the whole context in square brackets before adding the message at the end. e.g. if the
    logging metadata was `("%(asctime)s", "%(levelname)s", "%(name)s")`, the formatter would format log messages as e.g.
    `[2021-06-29 11:58:10,985 | INFO | octue.runner] This is a log message.`

    :param iter(str) logging_metadata: an iterable of context fields to use as context for every log message that the formatter is applied to
    :return logging.Formatter:
    """
    return logging.Formatter("[" + " | ".join(logging_metadata) + "]" + " %(message)s")


# Logging format for analysis runs. All handlers should use this logging format, to make logs consistently parseable
LOGGING_METADATA_WITH_TIMESTAMP = ("%(asctime)s", "%(levelname)s", "%(name)s")
FORMATTER_WITH_TIMESTAMP = create_formatter(LOGGING_METADATA_WITH_TIMESTAMP)
FORMATTER_WITHOUT_TIMESTAMP = create_formatter(LOGGING_METADATA_WITH_TIMESTAMP[1:])


def apply_log_handler(logger_name=None, handler=None, log_level=logging.INFO, formatter=None):
    """Apply a log handler with the given formatter to the logger with the given name.

    :param str|None logger_name: if this is `None`, the root logger is used
    :param logging.Handler handler: The handler to use. If None, default console handler will be attached.
    :param int log_level: ignore log messages below this level
    :param logging.Formatter|None formatter: if this is `None`, the default `FORMATTER_WITH_TIMESTAMP` is used
    :return logging.Logger:
    """
    handler = handler or logging.StreamHandler()
    handler.setFormatter(formatter or FORMATTER_WITH_TIMESTAMP)
    handler.setLevel(log_level)
    logger = logging.getLogger(name=logger_name)
    logger.addHandler(handler)
    logger.setLevel(log_level)

    for handler in logger.handlers:
        if type(handler).__name__ == "SocketHandler":
            # Log locally that a remote logger will be used.
            local_logger = logging.getLogger(__name__)
            temporary_handler = logging.StreamHandler()
            temporary_handler.setFormatter(formatter or FORMATTER_WITH_TIMESTAMP)
            temporary_handler.setLevel(log_level)
            local_logger.addHandler(temporary_handler)
            local_logger.setLevel(log_level)
            local_logger.info("Logs streaming to %s:%s", handler.host, str(handler.port))
            local_logger.removeHandler(temporary_handler)
            break

    return logger


def get_remote_handler(logger_uri, formatter=None):
    """Get a log handler for streaming logs to a remote URI accessed via HTTP or HTTPS. The given formatter is applied.

    :param str logger_uri: the URI to stream the logs to
    :param logging.Formatter|None formatter: if this is `None`, the `FORMATTER_WITH_TIMESTAMP` formatter is used
    :return logging.Handler:
    """
    parsed_uri = urlparse(logger_uri)

    if parsed_uri.scheme not in {"ws", "wss"}:
        raise ValueError(
            f"Only WS and WSS protocols currently supported for remote logger URI. Received {logger_uri!r}."
        )

    handler = logging.handlers.SocketHandler(host=parsed_uri.hostname, port=parsed_uri.port)
    handler.setFormatter(formatter or FORMATTER_WITH_TIMESTAMP)
    return handler



import uuid

from octue.exceptions import InvalidInputException
from octue.utils import gen_uuid


class Identifiable:
    """Mixin to allow instantiation of a class with a given uuid, or generate one on instantiation

    Prevents setting an id after an object is instantiated.

    Provides a basic str() method which will be overloaded by most Resource classes

    ```
    class MyResource(Identifiable):
        pass

    MyResource().id  # Some generated uuid
    MyResource(id='not_a_uuid')  # Raises exception
    MyResource(id='a10603a0-194c-40d0-a7b7-fcf9952c3690').id  # That same uuid
    ```
    """

    def __init__(self, *args, id=None, name=None, **kwargs):
        """Constructor for Identifiable class"""
        self._name = name
        super().__init__(*args, **kwargs)
        self._set_id(id)

    def __str__(self):
        return f"{self.__class__.__name__} {self._id}"

    def __repr__(self):
        return self.__str__()

    @property
    def id(self):
        return self._id

    @property
    def name(self):
        return self._name

    def _set_id(self, value):
        """Set the ID to the given value.

        :param str|uuid.UUID|None value:
        :return None:
        """
        # Store a boolean record of whether this object was created with a previously-existing uuid or was created new.
        self._created = True if value is None else False

        if isinstance(value, uuid.UUID):
            # If it's a uuid, stringify it
            value = str(value)

        elif isinstance(value, str):
            # If it's a string (or something similar which can be converted to UUID) check it's valid
            try:
                value = str(uuid.UUID(value))
            except ValueError:
                raise InvalidInputException(
                    f"Value of id '{value}' is not a valid uuid string or instance of class UUID"
                )

        elif value is not None:
            raise InvalidInputException(
                f"Value of id '{value}' must be a valid uuid string, an instance of class UUID or None"
            )

        self._id = value or gen_uuid()



import importlib
import logging
import os
import sys
import google.api_core.exceptions
from google.cloud import secretmanager

from octue.cloud.credentials import GCPCredentialsManager
from octue.log_handlers import apply_log_handler
from octue.resources import Child
from octue.resources.analysis import CLASS_MAP, Analysis
from octue.utils import gen_uuid
from twined import Twine


logger = logging.getLogger(__name__)


class Runner:
    """Runs analyses in the app framework

    The Runner class provides a set of configuration parameters for use by your application, together with a range of
    methods for managing input and output file parsing as well as controlling logging.

    :param Union[AppFrom, callable, str] app_src: Either an instance of the AppFrom manager class which has a run() method, or a function which accepts a single parameter (the instantiated analysis), or a string pointing to an application folder (which should contain an 'app.py' function like the templates)
    :param str|twined.Twine twine: path to the twine file, a string containing valid twine json, or a Twine instance
    :param str|dict|_io.TextIOWrapper|None configuration_values: The strand data. Can be expressed as a string path of a *.json file (relative or absolute), as an open file-like object (containing json data), as a string of json data or as an already-parsed dict.
    :param str|dict|_io.TextIOWrapper|None configuration_manifest: The strand data. Can be expressed as a string path of a *.json file (relative or absolute), as an open file-like object (containing json data), as a string of json data or as an already-parsed dict.
    :param Union[str, path-like, None] output_manifest_path: Path where output data will be written
    :param Union[str, dict, None] children: The children strand data. Can be expressed as a string path of a *.json file (relative or absolute), as an open file-like object (containing json data), as a string of json data or as an already-parsed dict.
    :param bool skip_checks: If true, skip the check that all files in the manifest are present on disc - this can be an extremely long process for large datasets.
    :param str|None project_name: name of Google Cloud project to get credentials from
    :return None:
    """

    def __init__(
        self,
        app_src,
        twine="twine.json",
        configuration_values=None,
        configuration_manifest=None,
        output_manifest_path=None,
        children=None,
        skip_checks=False,
        project_name=None,
    ):
        self.app_src = app_src
        self.output_manifest_path = output_manifest_path
        self.children = children
        self.skip_checks = skip_checks

        # Ensure the twine is present and instantiate it.
        if isinstance(twine, Twine):
            self.twine = twine
        else:
            self.twine = Twine(source=twine)

        logger.debug("Parsed twine with strands %r", self.twine.available_strands)

        # Validate and initialise configuration data
        self.configuration = self.twine.validate(
            configuration_values=configuration_values,
            configuration_manifest=configuration_manifest,
            cls=CLASS_MAP,
        )
        logger.debug("Configuration validated.")

        # Set path for configuration manifest.
        # TODO this is hacky, we need to rearchitect the twined validation so we can do this kind of thing in there
        self.configuration["configuration_manifest"] = self._update_manifest_path(
            self.configuration.get("configuration_manifest", None),
            configuration_manifest,
        )

        self._project_name = project_name

    def run(
        self,
        analysis_id=None,
        input_values=None,
        input_manifest=None,
        analysis_log_level=logging.INFO,
        analysis_log_handler=None,
    ):
        """Run an analysis

        :param str|None analysis_id: UUID of analysis
        :param Union[str, dict, None] input_values: the input_values strand data. Can be expressed as a string path of a *.json file (relative or absolute), as an open file-like object (containing json data), as a string of json data or as an already-parsed dict.
        :param Union[str, octue.resources.manifest.Manifest, None] input_manifest: The input_manifest strand data. Can be expressed as a string path of a *.json file (relative or absolute), as an open file-like object (containing json data), as a string of json data or as an already-parsed dict.
        :param str analysis_log_level: the level below which to ignore log messages
        :param logging.Handler|None analysis_log_handler: the logging.Handler instance which will be used to handle logs for this analysis run. Handlers can be created as per the logging cookbook https://docs.python.org/3/howto/logging-cookbook.html but should use the format defined above in LOG_FORMAT.
        :return: None
        """
        if hasattr(self.twine, "credentials"):
            self._populate_environment_with_google_cloud_secrets()
            credentials = self.twine.credentials
        else:
            credentials = None

        inputs = self.twine.validate(
            input_values=input_values,
            input_manifest=input_manifest,
            credentials=credentials,
            children=self.children,
            cls=CLASS_MAP,
            allow_missing=False,
            allow_extra=False,
        )
        logger.debug("Inputs validated.")

        if inputs["children"] is not None:
            inputs["children"] = {
                child["key"]: Child(name=child["key"], id=child["id"], backend=child["backend"])
                for child in inputs["children"]
            }

        # TODO this is hacky, we need to rearchitect the twined validation so we can do this kind of thing in there
        inputs["input_manifest"] = self._update_manifest_path(
            inputs.get("input_manifest", None),
            input_manifest,
        )

        outputs_and_monitors = self.twine.prepare("monitors", "output_values", "output_manifest", cls=CLASS_MAP)

        # TODO this is hacky, we need to rearchitect the twined validation so we can do this kind of thing in there
        outputs_and_monitors["output_manifest"] = self._update_manifest_path(
            outputs_and_monitors.get("output_manifest", None),
            self.output_manifest_path,
        )

        analysis_id = str(analysis_id) if analysis_id else gen_uuid()
        analysis_logger_name = f"{__name__} | analysis-{analysis_id}"

        # Apply the default stderr log handler to the analysis logger.
        analysis_logger = apply_log_handler(logger_name=analysis_logger_name, log_level=analysis_log_level)

        # Also apply the given analysis log handler if given.
        if analysis_log_handler:
            apply_log_handler(
                logger_name=analysis_logger_name, handler=analysis_log_handler, log_level=analysis_log_level
            )

        # Stop messages logged by the analysis logger being repeated by the root logger.
        analysis_logger.propagate = False

        analysis = Analysis(
            id=analysis_id,
            logger=analysis_logger,
            twine=self.twine,
            skip_checks=self.skip_checks,
            **self.configuration,
            **inputs,
            **outputs_and_monitors,
        )

        try:
            if hasattr(self.app_src, "run"):
                self.app_src.run(analysis)
            elif isinstance(self.app_src, str):
                with AppFrom(self.app_src) as app:
                    app.run(analysis)
            else:
                self.app_src(analysis)

        except ModuleNotFoundError as e:
            raise ModuleNotFoundError(f"{e.msg} in {os.path.abspath(self.app_src)!r}.")

        except Exception as e:
            analysis_logger.error(str(e))
            raise e

        return analysis

    @staticmethod
    def _update_manifest_path(manifest, pathname):
        """A Quick hack to stitch the new Pathable functionality in the 0.1.4 release into the CLI and runner.

        The way we define a manifest path can be more robustly implemented as we migrate functionality into the twined
        library

        :param manifest:
        :type manifest:
        :param pathname:
        :type pathname:
        :return:
        :rtype:
        """
        if manifest is not None and hasattr(pathname, "endswith"):
            if pathname.endswith(".json"):
                manifest.path = os.path.split(pathname)[0]

        # Otherwise do nothing and rely on manifest having its path variable set already
        return manifest

    def _populate_environment_with_google_cloud_secrets(self):
        """Get any secrets specified in the credentials strand from Google Cloud Secret Manager and put them in the
        local environment, ready for use by the runner.

        :return None:
        """
        missing_credentials = tuple(
            credential for credential in self.twine.credentials if credential["name"] not in os.environ
        )

        if not missing_credentials:
            return

        google_cloud_credentials = GCPCredentialsManager().get_credentials()
        secrets_client = secretmanager.SecretManagerServiceClient(credentials=google_cloud_credentials)

        if google_cloud_credentials is None:
            project_name = self._project_name
        else:
            project_name = google_cloud_credentials.project_id

        for credential in missing_credentials:
            secret_path = secrets_client.secret_version_path(
                project=project_name, secret=credential["name"], secret_version="latest"
            )

            try:
                secret = secrets_client.access_secret_version(name=secret_path).payload.data.decode("UTF-8")
            except google.api_core.exceptions.NotFound:
                # No need to raise an error here as the Twine validation that follows will do so.
                continue

            os.environ[credential["name"]] = secret


def unwrap(fcn):
    """Recurse through wrapping to get the raw function without decorators."""
    if hasattr(fcn, "__wrapped__"):
        return unwrap(fcn.__wrapped__)
    return fcn


class AppFrom:
    """Context manager that imports module 'app' from user's code base at a location app_path.

     The manager will issue a warning if an existing module called "app" is already loaded.

     The manager makes a temporary addition to the system path (to ensure app is loaded from the correct path)

     The manager will unload the module (by deleting it from sys.modules) on exit, enabling

    with AppFrom('/path/to/dir') as app:
        Runner().run(app)

    """

    def __init__(self, app_path="."):
        self.app_path = os.path.abspath(os.path.normpath(app_path))
        logger.debug("Initialising AppFrom context at app_path %s", self.app_path)
        self.app_module = None

    def __enter__(self):
        # Warn on an app present on the system path
        if "app" in sys.modules.keys():
            logger.warning(
                "Module 'app' already on system path. Using 'AppFrom' context will yield unexpected results. Avoid using 'app' as a python module, except for your main entrypoint"
            )

        # Insert the present directory first on the system path
        sys.path.insert(0, self.app_path)

        # Import the app from the present directory
        self.app_module = importlib.import_module("app")

        # Immediately clean up the entry to the system path (don't use "remove" because if the user has it in their
        # path, this'll be an unexpected side effect, and don't do it in cleanup in case the called code inserts a path)
        sys.path.pop(0)
        logger.debug("Imported app at app_path and cleaned up temporary modification to sys.path %s", self.app_path)

        return self

    def __exit__(self, exc_type, exc_value, traceback):
        # Unload the imported module
        del sys.modules["app"]
        logger.debug("Deleted app from sys.modules and cleaned up (app_path %s)", self.app_path)

    @property
    def run(self):
        """Returns the unwrapped run function from app.py in the application's root directory"""
        return unwrap(self.app_module.run)



import json
from json import JSONDecoder
import dateutil.parser


class OctueJSONDecoder(JSONDecoder):
    """A JSON Decoder to convert default json objects into their Datafile, Dataset or Manifest classes as appropriate"""

    def __init__(self, *args, object_hook=None, **kwargs):
        json.JSONDecoder.__init__(self, object_hook=object_hook or self.object_hook, *args, **kwargs)

    def object_hook(self, obj):
        if "_type" not in obj:
            return obj

        if obj["_type"] == "set":
            return set(obj["items"])

        if obj["_type"] == "datetime":
            return dateutil.parser.parse(obj["value"])

        return obj
